"""Files in this directory are generated by running:

    python -m dagster_cloud.pex.grpc.compile
"""

from . import multi_pex_api_pb2
from .multi_pex_api_pb2_grpc import MultiPexApiServicer, MultiPexApiStub



import os
import sys
import threading
import time
from concurrent.futures import ThreadPoolExecutor

import grpc
from dagster._grpc.__generated__.api_pb2_grpc import (
    DagsterApiServicer,
    add_DagsterApiServicer_to_server,
)
from dagster._grpc.server import server_termination_target
from dagster._grpc.utils import max_rx_bytes, max_send_bytes
from dagster._serdes import deserialize_as, serialize_dagster_namedtuple
from dagster._utils.error import serializable_error_info_from_exc_info
from grpc_health.v1 import health, health_pb2, health_pb2_grpc

from ..manager import MultiPexManager
from .__generated__ import multi_pex_api_pb2
from .__generated__.multi_pex_api_pb2_grpc import (
    MultiPexApiServicer,
    add_MultiPexApiServicer_to_server,
)
from .types import (
    CreatePexServerArgs,
    CreatePexServerResponse,
    GetPexServersArgs,
    GetPexServersResponse,
    PexServerHandle,
    ShutdownPexServerArgs,
    ShutdownPexServerResponse,
)


class MultiPexApiServer(MultiPexApiServicer):
    def __init__(
        self,
        pex_manager: MultiPexManager,
    ):
        self._pex_manager = pex_manager

    def CreatePexServer(self, request, _context):
        create_pex_server_args = deserialize_as(request.create_pex_server_args, CreatePexServerArgs)
        try:
            self._pex_manager.create_pex_server(
                create_pex_server_args.server_handle,
                create_pex_server_args.code_deployment_metadata,
            )
            response = CreatePexServerResponse()
        except:
            response = serializable_error_info_from_exc_info(sys.exc_info())

        return multi_pex_api_pb2.CreatePexServerReply(
            create_pex_server_response=serialize_dagster_namedtuple(response)
        )

    def GetPexServers(self, request, _context):
        get_pex_servers_args = deserialize_as(request.get_pex_servers_args, GetPexServersArgs)
        try:
            pex_server_handles = self._pex_manager.get_pex_servers(
                get_pex_servers_args.deployment_name,
                get_pex_servers_args.location_name,
            )
            response = GetPexServersResponse(server_handles=pex_server_handles)
        except:
            response = serializable_error_info_from_exc_info(sys.exc_info())

        return multi_pex_api_pb2.GetPexServersReply(
            get_pex_servers_response=serialize_dagster_namedtuple(response)
        )

    def ShutdownPexServer(self, request, _context):
        shutdown_pex_server_args = deserialize_as(
            request.shutdown_pex_server_args, ShutdownPexServerArgs
        )
        try:
            self._pex_manager.shutdown_pex_server(shutdown_pex_server_args.server_handle)
            response = ShutdownPexServerResponse()
        except:
            response = serializable_error_info_from_exc_info(sys.exc_info())
        return multi_pex_api_pb2.ShutdownPexServerReply(
            shutdown_pex_server_response=serialize_dagster_namedtuple(response)
        )

    def Ping(self, request, _context):
        echo = request.echo
        return multi_pex_api_pb2.PingReply(echo=echo)


class DagsterPexProxyApiServer(DagsterApiServicer):
    def __init__(self, pex_manager: MultiPexManager):
        self._pex_manager = pex_manager

    def _get_handle_from_metadata(self, context) -> PexServerHandle:
        metadict = dict(context.invocation_metadata())

        if "deployment" not in metadict:
            raise Exception("missing `deployment` in request metadata")

        if "location" not in metadict:
            raise Exception("missing `location` in request metadata")

        if "timestamp" not in metadict:
            raise Exception("missing `timestamp` in request metadata")

        return PexServerHandle(
            deployment_name=metadict["deployment"],
            location_name=metadict["location"],
            metadata_update_timestamp=int(metadict["timestamp"]),
        )

    def _query(self, api_name: str, request, context):
        return self._pex_manager.get_pex_grpc_client(  # pylint:disable=protected-access
            self._get_handle_from_metadata(context)
        )._get_response(api_name, request)

    def _streaming_query(self, api_name: str, request, context):
        return self._pex_manager.get_pex_grpc_client(  # pylint:disable=protected-access
            self._get_handle_from_metadata(context)
        )._get_streaming_response(api_name, request)

    def ExecutionPlanSnapshot(self, request, context):
        return self._query("ExecutionPlanSnapshot", request, context)

    def ListRepositories(self, request, context):
        return self._query("ListRepositories", request, context)

    def Ping(self, request, context):
        return self._query("Ping", request, context)

    def GetServerId(self, request, context):
        return self._query("GetServerId", request, context)

    def GetCurrentImage(self, request, context):
        return self._query("GetCurrentImage", request, context)

    def StreamingExternalRepository(self, request, context):
        return self._streaming_query("StreamingExternalRepository", request, context)

    def Heartbeat(self, request, context):
        return self._query("Heartbeat", request, context)

    def StreamingPing(self, request, context):
        return self._streaming_query("StreamingPing", request, context)

    def ExternalPartitionNames(self, request, context):
        return self._query("ExternalPartitionNames", request, context)

    def ExternalNotebookData(self, request, context):
        return self._query("ExternalNotebookData", request, context)

    def ExternalPartitionConfig(self, request, context):
        return self._query("ExternalPartitionConfig", request, context)

    def ExternalPartitionTags(self, request, context):
        return self._query("ExternalPartitionTags", request, context)

    def ExternalPartitionSetExecutionParams(self, request, context):
        return self._query("ExternalPartitionSetExecutionParams", request, context)

    def ExternalPipelineSubsetSnapshot(self, request, context):
        return self._query("ExternalPipelineSubsetSnapshot", request, context)

    def ExternalRepository(self, request, context):
        return self._query("ExternalRepository", request, context)

    def ExternalJob(self, request, context):
        return self._query("ExternalJob", request, context)

    def ExternalScheduleExecution(self, request, context):
        return self._streaming_query("ExternalScheduleExecution", request, context)

    def ExternalSensorExecution(self, request, context):
        return self._streaming_query("ExternalSensorExecution", request, context)

    def ShutdownServer(self, request, context):
        return self._query("ShutdownServer", request, context)

    def CancelExecution(self, request, context):
        return self._query("CancelExecution", request, context)

    def CanCancelExecution(self, request, context):
        return self._query("CanCancelExecution", request, context)

    def StartRun(self, request, context):
        return self._query("StartRun", request, context)


def run_multipex_server(port, print_fn, host="localhost", max_workers=None):
    server = grpc.server(
        ThreadPoolExecutor(max_workers=max_workers),
        compression=grpc.Compression.Gzip,
        options=[
            ("grpc.max_send_message_length", max_send_bytes()),
            ("grpc.max_receive_message_length", max_rx_bytes()),
        ],
    )

    with MultiPexManager() as pex_manager:
        pex_api_servicer = MultiPexApiServer(
            pex_manager=pex_manager,
        )

        server_termination_event = threading.Event()

        dagster_api_servicer = DagsterPexProxyApiServer(pex_manager=pex_manager)
        health_servicer = health.HealthServicer()

        health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)

        add_MultiPexApiServicer_to_server(pex_api_servicer, server)
        add_DagsterApiServicer_to_server(dagster_api_servicer, server)

        server_address = host + ":" + str(port)
        res = server.add_insecure_port(server_address)
        if res != port:
            raise Exception(f"Could not bind to port {port}")

        server_desc = f"Pex server on port {port} in process {os.getpid()}"

        server.start()

        print_fn(f"Started {server_desc}")

        health_servicer.set("MultiPexApi", health_pb2.HealthCheckResponse.SERVING)

        server_termination_thread = threading.Thread(
            target=server_termination_target,
            args=[server_termination_event, server],
            name="grpc-server-termination",
        )
        server_termination_thread.daemon = True
        server_termination_thread.start()

        server.wait_for_termination()

        print_fn(f"Shutting down {server_desc}")
        server_termination_thread.join()
        pex_api_servicer.cleanup()
        dagster_api_servicer.cleanup()

    print_fn("Server shut down.")


def wait_for_grpc_server(client, timeout=180):
    start_time = time.time()

    while True:
        try:
            client.ping("")
            return
        except Exception:
            last_error = serializable_error_info_from_exc_info(sys.exc_info())
            print(str(last_error))

        if time.time() - start_time > timeout:
            raise Exception(
                f"Timed out after waiting {timeout}s for server. "
                f"Most recent connection error: {str(last_error)}"
            )

        time.sleep(1)



import os
import threading
from contextlib import AbstractContextManager
from typing import Dict, List, NamedTuple

from dagster import _check as check
from dagster._core.errors import DagsterUserCodeUnreachableError
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._grpc.client import DagsterGrpcClient, client_heartbeat_thread
from dagster._grpc.server import GrpcServerProcess
from dagster_cloud_cli.core.workspace import CodeDeploymentMetadata, PexMetadata

from .grpc.types import PexServerHandle


class PexProcessEntry(
    NamedTuple(
        "_PexProcessEntry",
        [
            ("pex_server_handle", PexServerHandle),
            ("grpc_server_process", GrpcServerProcess),
            ("grpc_client", DagsterGrpcClient),
            ("heartbeat_shutdown_event", threading.Event),
            ("heartbeat_thread", threading.Thread),
        ],
    )
):
    def __new__(
        cls,
        pex_server_handle: PexServerHandle,
        grpc_server_process: GrpcServerProcess,
        grpc_client: DagsterGrpcClient,
        heartbeat_shutdown_event: threading.Event,
        heartbeat_thread: threading.Thread,
    ):
        return super(PexProcessEntry, cls).__new__(
            cls,
            check.inst_param(pex_server_handle, "pex_server_handle", PexServerHandle),
            check.inst_param(grpc_server_process, "grpc_server_process", GrpcServerProcess),
            check.inst_param(grpc_client, "grpc_client", DagsterGrpcClient),
            check.inst_param(heartbeat_shutdown_event, "heartbeat_shutdown_event", threading.Event),
            check.inst_param(heartbeat_thread, "heartbeat_thread", threading.Thread),
        )


class MultiPexManager(AbstractContextManager):
    def __init__(
        self,
    ):
        # Keyed by hash of PexServerHandle
        self._pex_servers: Dict[str, PexProcessEntry] = {}
        self._pex_servers_lock = threading.Lock()
        self._heartbeat_ttl = 60

    def download_pex_from_s3(self, _pex_metadata: PexMetadata):
        # TODO Implement for real
        return "./source.pex"

    def get_pex_grpc_client(self, server_handle: PexServerHandle):
        handle_id = server_handle.get_id()
        with self._pex_servers_lock:
            if not handle_id in self._pex_servers:
                raise Exception("No server created with the given handle")

            return self._pex_servers[handle_id].grpc_client

    def get_pex_servers(self, deployment_name, location_name: str) -> List[PexServerHandle]:
        with self._pex_servers_lock:
            return [
                server.pex_server_handle
                for server in self._pex_servers.values()
                if server.pex_server_handle.deployment_name == deployment_name
                and server.pex_server_handle.location_name == location_name
            ]

    def create_pex_server(
        self,
        server_handle: PexServerHandle,
        code_deployment_metadata: CodeDeploymentMetadata,
    ):
        print(
            f"Creating new pex server for {server_handle.deployment_name}:{server_handle.location_name}"
        )

        executable_path = self.download_pex_from_s3(
            check.not_none(code_deployment_metadata.pex_metadata)
        )

        metadata = code_deployment_metadata

        loadable_target_origin = LoadableTargetOrigin(
            executable_path=executable_path,
            python_file=metadata.python_file,
            package_name=metadata.package_name,
            module_name=metadata.module_name,
            working_directory=metadata.working_directory,
            attribute=metadata.attribute,
        )

        server_process = GrpcServerProcess(
            loadable_target_origin=loadable_target_origin,
            heartbeat=True,
            heartbeat_timeout=self._heartbeat_ttl,
            startup_timeout=0,  # don't wait for startup, agent will poll for status
            log_level="INFO",
            # Temporary for testing
            env={**os.environ.copy(), "PEX_PATH": os.path.join(os.getcwd(), "deps.pex")},
        )

        client = DagsterGrpcClient(
            port=server_process.port,
            socket=server_process.socket,
            host="localhost",
            use_ssl=False,
        )

        heartbeat_shutdown_event = threading.Event()
        heartbeat_thread = threading.Thread(
            target=client_heartbeat_thread,
            args=(client, heartbeat_shutdown_event),
        )
        heartbeat_thread.daemon = True
        heartbeat_thread.start()

        # open question - how do we know when to spin down old pexes / how do you reload a pex
        # in a zero-downtime way (very similar questions to the process agent really)
        with self._pex_servers_lock:
            self._pex_servers[server_handle.get_id()] = PexProcessEntry(
                pex_server_handle=server_handle,
                grpc_server_process=server_process,
                grpc_client=client,
                heartbeat_shutdown_event=heartbeat_shutdown_event,
                heartbeat_thread=heartbeat_thread,
            )

    def shutdown_pex_server(self, server_handle: PexServerHandle):
        handle_id = server_handle.get_id()
        pex_server = None
        with self._pex_servers_lock:
            pex_server = self._pex_servers.get(handle_id)
            if pex_server:
                del self._pex_servers[handle_id]

        if pex_server:
            pex_server.heartbeat_shutdown_event.set()
            pex_server.heartbeat_thread.join()
            try:
                pex_server.grpc_client.shutdown_server()
            except DagsterUserCodeUnreachableError:
                # Server already shutdown
                pass
            pex_server.grpc_server_process.wait()

    def __exit__(self, exception_type, exception_value, traceback):
        for _handle, pex_server in self._pex_servers.items():
            pex_server.heartbeat_shutdown_event.set()
            pex_server.heartbeat_thread.join()

        for _handle, pex_server in self._pex_servers.items():
            try:
                pex_server.grpc_client.shutdown_server()
            except DagsterUserCodeUnreachableError:
                # Server already shutdown
                pass
            pex_server.grpc_server_process.wait()



import logging
import os
from typing import Optional

from dagster._core.utils import coerce_valid_log_level
from typer import Option, Typer, echo

from ..grpc.server import run_multipex_server

app = Typer(hidden=True)


@app.command(short_help="Run multi-pex server to spin up subprocesses via PEX files")
def grpc(
    host: str = Option(default="localhost"),
    port: int = Option(default=...),
    max_workers: Optional[int] = Option(default=None),
):
    run_multipex_server(port=port, host=host, print_fn=echo, max_workers=max_workers)
